.. -*- rst -*-

.. _versions-call:

API versions
************


List API versions
~~~~~~~~~~~~~~~~~

.. rest_method::  GET /versions

Lists information about all Image service API versions supported by this
deployment, including the URIs.


Normal response codes: 200


Request
-------

There are no request parameters.



Response Example
----------------

.. literalinclude:: samples/image-versions-response.json
   :language: json


List API versions
~~~~~~~~~~~~~~~~~

.. rest_method::  GET /

Lists information about all Image service API versions supported by this
deployment, including the URIs.



Normal response codes: 300


Request
-------

There are no request parameters.


Response Example
----------------

.. literalinclude:: samples/image-versions-response.json
   :language: json
